from pwn import *
import hashlib
import requests
import zlib
from Crypto.Cipher import ChaCha20_Poly1305
import random
from Crypto.Util.number import *
context.log_level = 'critical'

r = remote('dhcppp.chal.pwni.ng',1337)
#r = process('./server.py')

data = r.recvline().decode()[:-1]
data = data.split(' ')
flag_mac = bytes.fromhex("53 79 82 b5 97 eb")
dhcp_mac = bytes.fromhex("1b 7d 6f 49 37 c9")
whatever_mac = bytes.fromhex("00 00 00 00 00 00")

def calc_crc(msg):
    return zlib.crc32(msg).to_bytes(4, "little")

P = 2**130- 5

def get_blocks(c):
    c = c + bytes.fromhex('000000000000000000002e00000000000000')
    real_c = []
    for i in range(0, len(c), 16):
        block = c[i:i+16]
        block = block + b'\x01'
        assert len(block) == 17
        real_c.append(int.from_bytes(block, byteorder='little'))

    return real_c

def tag(c, r, s):
    mod2 = 2**128
    real_c = get_blocks(c)
    c = real_c
    n = len(c)

    res = 0
    for i in range(n):
        res += c[i]
        res = (r*res) % P
    return (res + s) % mod2

def atk(pair1, pair2, nonce, m3):
    m1, c1, t1 = pair1
    m2, c2, t2 = pair2

    t1 = int.from_bytes(t1, byteorder='little')
    t2 = int.from_bytes(t2, byteorder='little')
    keystream = bytes(a ^ b for a, b in zip(c1, m1))
    c3 = bytes(a ^ b for a, b in zip(m3, keystream))

    for k1 in range(5):
        for k2 in range(5):
            t1_ = t1 + k1*2**128
            t2_ = t2 + k2*2**128
            deltat1 = (t2_-t1_)
            for i in range(4):
                guess = pow((deltat1 * inverse(K, P)), inverse(i, P-1), P)

                if guess & 0x0ffffffc0ffffffc0ffffffc0fffffff == guess and guess != 1:
                    r = guess
                    print('Found r:', r)
                    s = tag(c1, r, 0) - t1 
                    t3 = tag(c3, r, s)
                    t3 = long_to_bytes(t3)[::-1]
                    return c3, t3

    assert False, "Attack failed"

def dhcp_msg(dev_name):
    r.recvuntil(b'> ')
    msg = whatever_mac + dhcp_mac + b'\x01' + dev_name
    r.sendline(msg.hex().encode())
    return bytes.fromhex(r.recvline().rstrip().decode())

NUM_LEASES = 64 - 2

for i in range(NUM_LEASES - 2): # 2 leases were already taken
    ret =  dhcp_msg(b'lolada' + str(i).encode())

m1 = b'AAAAAAAAAAAAAAAAAAAAAAAAA'
L1 = dhcp_msg(m1)


for i in range(NUM_LEASES - 1): # Rotate all leases
    ret =  dhcp_msg(b'loladaa' + str(i).encode())

m2 = b'AAAAAAAAAAAAAAAAAAAAAAAAB'
L2 = dhcp_msg(m2) # Almost equal (so nonce is equal) but differ in last bytes

m1 = bytes.fromhex('c0a80102c0a80101ffffff0008080808080804044141414141414141414141414141414141414141414141414100')
m2 = bytes.fromhex('c0a80102c0a80101ffffff0008080808080804044141414141414141414141414141414141414141414141414200')

def extract(lease):
    src_mac = lease[:6]
    assert src_mac == dhcp_mac
    dst_mac = lease[6:12]
    assert dst_mac == whatever_mac
    msg = lease[12:]
    assert msg.startswith(b'\x02')
    pkt = msg[1:-4]
    ct = pkt[:-28]
    tag = pkt[-28:-12]
    nonce = pkt[-12:]
    return ct, tag, nonce

c1, t1, n1 = extract(L1)
c2, t2, n2 = extract(L2)


#K = 237684487542793012780631851008
print('blocks1:', get_blocks(c1))
print('blocks2:', get_blocks(c2))
K = get_blocks(c2)[2] - get_blocks(c1)[2] # Difference of c's
print('K=', K)

assert n1 == n2 # must use the same nonce for the attack to work
n = n1

# Some debug prints
print('m1=', m1.hex())
print('c1=', c1.hex())
print('t1=', t1.hex())
print('m2=', m2.hex())
print('c2=', c2.hex())
print('t2=', t2.hex())
print('nonce=', n.hex())

dns1 = '207.154.233.177' # Our DNS server
dns2 = dns1
dns1 = bytes([int(x) for x in dns1.split(".")])
dns2 = bytes([int(x) for x in dns2.split(".")])

m3 = b'\xc0\xa8\x01\x02\xc0\xa8\x01\x01\xff\xff\xff\x00' + dns1 + dns2 + b'AAAAAAAAAAAAAAAAAAAAAAAAB\x00'

c3, t3 = atk((m1, c1, t1), (m2, c2, t2), n, m3)
forged_lease  = b'\x02' + c3 + t3 + n + calc_crc(m3)

flagserver_pkt = whatever_mac + flag_mac + forged_lease

r.sendline(flagserver_pkt.hex().encode())
r.sendline(b'0300') # Get the flag

while 1:
    try:
        print(r.recvline())
    except:break